#! /usr/bin/env tclsh

# Convert a NH or NHX tree to treecon format
# version 0.1.20060515
# start with one garbage option to get usage information
# Jan Wuyts (jan.wuyts at psb.ugent.be)

set ADD_NODENR_AS_BOOTSTRAP 0
set DUMP_BOOTSTRAPS 0
set USE_INFILE 0
set USE_OUTFILE 0
set RESET_DISTANCE 0
set SCALE_DISTANCE 0
set SLIDE_ROOT 0
set SLIDE_FACTOR 0

proc giveusage {} {
	puts stderr "nhx2treecon.tcl: convert nh or nhx tree file to treecon for windows format"
	puts stderr "\tdefault is to accept tree from stdin and write treeconfile to stdout"
	puts stderr "\n\tOptions:"
	puts stderr "\t-i <filename> : use <filename> as Input"
	puts stderr "\t-o <filename> : use <filename> as Output"
	puts stderr "\t-b : use internal node numbers as Bootstrap values"
	puts stderr "\t-n : igNore bootstrap values even if they are present in the nhx tree"
	puts stderr "\t-d <distance (float)> : reset all Distances to <distance> (usefull if no distances are given in input tree)"
	puts stderr "\t-s <scale (float)> : Scale all distances by factor <scale>."
	puts stderr "\t-l <relative position> (0 < float < 1) : sLide root: distribute distance from root to outgroup between the 2 horizontal branches."
}

proc IsBifurcated tree { # recursive alghoritm 
	# Terminating condition 1
	if {[CountOrgs $tree] <= 2} {
		return 1
	}
	set parts [BreakParts $tree]
	# Terminating condition 2
	if {[llength $parts] > 2} {
		return 0
	}
	# Recursion
	if {[IsBifurcated [lindex $parts 0]] && [IsBifurcated [lindex $parts 1]]} {
		return 1
	} else {
		return 0
	}
}

proc BreakParts tree {
# Give this a bracket format tree and it will return you a list of the first child nodes
	set SEPCHAR {,}     ; #
	set OPENBR {(}      ; #
	set CLOSEBR {)}     ; # Also change regexp in first if condition of while loop !
	set OPENCOMMENT {[} ; #
	set CLOSECOMMENT {]}; #
	set length [string length $tree]
	set pos 0
	set parts [list]
	set count 0
	set incomment 0
	while {$pos < $length} {
		set char [string index $tree $pos]
		if [regexp {[^,()\[\]]} $char] { # Speed things up a bit
			incr pos
			continue
		} elseif {$char == $OPENCOMMENT}  {
			incr incomment
		} elseif {$char == $CLOSECOMMENT} {
			incr incomment -1
		} elseif {!$incomment && ($char == $OPENBR)} {
			if {$count == 0} {
				set prev [expr $pos + 1]
			}
			incr count
		} elseif {!$incomment && ($char == $CLOSEBR)} {
			incr count -1
			if {$count == 0} {
				set tmp [string range $tree $prev [expr $pos - 1] ]
				regsub {^\s+|\s+$} $tmp "" tmp; # remove empty space from begin/end
				lappend parts $tmp
				set prev [expr $pos + 1]
			}
		} elseif {!$incomment && ($char == $SEPCHAR) && ($count == 1)} {
			set tmp [string range $tree $prev [expr $pos - 1] ]
			regsub {^\s+|\s+$} $tmp "" tmp; # remove empty space from begin/end
			lappend parts $tmp
			set prev [expr $pos + 1]
		}
		incr pos
	}
	return $parts
}

proc CountOrgs tree {
# How many endnodes are in this tree/node
# Calculate as (#branches + 1)
	set SEPCHAR {,}
	set OPENCOMMENT {[}
	set CLOSECOMMENT {]}
	set length [string length $tree]
	set pos 0
	set count 1
	set incomment 0
	while {$pos < $length} {
		set char [string index $tree $pos]
		if       {$char == $OPENCOMMENT} {
			incr incomment
		} elseif {$char == $CLOSECOMMENT} {
			incr incomment -1
		} elseif {!$incomment && ($char == $SEPCHAR)} {
			incr count
		}
		incr pos
	}
	return $count
}

# tree: string: tree in NH(X) format
# treearray: array leave empty and it will be filled with:
#	key: node number
#	value: string containing sub-tree on that node
# nodenrsplitarray: array: leave empty and it will be filled with:
#	key: node number
#	value: list with 2 element representing the children of node $key
# parentarray: array: leave empty and it will be filled with:
#	key: node number
#	value: node number for parent of node $key
# sisterarray: array: leave empty and it will be filled with:
#	key: node number
#	value: node number of sister node of $key
# nodecounter: integer: start number for nodes: set at 0 and nodes will be numbered 1,2,3,... (1 is root node)
# endnodes: list: leave empty and it will be filled with the node numbers of the endnodes
proc DecomposeAndNumber {tree treearray nodenrsplitarray parentarray sisterarray nodecounter endnodes} {
	upvar $treearray trarr
	upvar $nodenrsplitarray nnrsplarr
	upvar $parentarray parent
	upvar $sisterarray sister
	upvar $nodecounter nodecnt
	upvar $endnodes endnodelist
	incr nodecnt
	set keep $nodecnt
	if {$tree != ""} {
		set trarr($nodecnt) $tree
		set parts [BreakParts $tree]
		if {$parts != ""} {
			set x [DecomposeAndNumber [lindex $parts 0] trarr nnrsplarr parent sister nodecnt endnodelist]
			set y [DecomposeAndNumber [lindex $parts 1] trarr nnrsplarr parent sister nodecnt endnodelist]
			set nnrsplarr($keep) [list $x $y]
			set parent($x) $keep
			set parent($y) $keep
			set sister($x) $y
			set sister($y) $x
		} else {
			lappend endnodelist $keep
		}
	}
	return $keep
}

# Use numbers of internal nodes as bootstrap values of these nodes
proc MakeNodeNrBootst {intnodes infoarray} {
	upvar $infoarray info
	foreach node $intnodes {
		if [info exists info($node,nn)] {
			set info($node,bs) $info($node,nn)
		} else {
			set info($node,bs) 0
		}
	}
}

proc ReadNodeInfo {treearray endnodes infoarray} {
	upvar $treearray trarr
	upvar $endnodes endnodelist
	upvar $infoarray infarray
	set x 1; set nrnodes [expr 2 * [llength $endnodelist]]
	set nodes [list]
	while {$x < $nrnodes} {
		lappend nodes $x
		incr x
	}
	foreach node $nodes {
#		puts "$node:\t$trarr($node)"
		if ![regexp {\)[\'\" ]*([0-9.]+)[\'\" ]*:?[0-9.]*[^()]*$} $trarr($node) tmp infarray($node,nn)] {
			set infarray($node,nn) 0
		}
		if ![regexp {\)?[0-9.\'\"]*\:([0-9.]+)[^()]*$} $trarr($node) tmp infarray($node,ln)] {
			set infarray($node,ln) 0.0
		}
		regexp {^([^:\[\]]+):?[0-9.]*(\[[^\]]\])?$} $trarr($node) tmp infarray($node,sn)
		if [regexp {\[([^\]]+)\]$} $trarr($node) tmp nhxstring] {
			set nhxstringparts [split $nhxstring ":"]
			foreach part [lrange $nhxstringparts 1 end] {
				set keyvalue [split $part "="]
				switch [lindex $keyvalue 0] {
					"B" {
						set infarray($node,bs) [string range $part 2 end]
					}
					"S" {
						set infarray($node,sp) [string range $part 2 end]
					}
					"T" {
						set infarray($node,tx) [string range $part 2 end]
					}
					"E" {
						set infarray($node,ec) [string range $part 2 end]
					}
					"D" {
						set infarray($node,du) [string range $part 2 end]
					}
					"O" {
						set infarray($node,or) [string range $part 2 end]
					}
					"SO" {
						set infarray($node,so) [string range $part 3 end]
					}
					"L" {
						set infarray($node,ll) [string range $part 2 end]
					}
					"Sw" {
						set infarray($node,sw) [string range $part 3 end]
					}
					"Co" {
						set infarray($node,co) [string range $part 3 end]
					}
				}
			}
		}
	}
}

proc WriteTreeconFile {intnodes extnodes children namearray infoarray} {
	upvar $infoarray info
	upvar $children childs
	upvar $namearray names
	set intnodes [lsort -integer -decreasing $intnodes]
	set extnodes [lsort -integer -increasing $extnodes]
	set x 1
	foreach node [concat $extnodes $intnodes] {
		set transnode($node) $x
		incr x
	}
	set bootstrap 0
	foreach node $intnodes {
		if [info exists info($node,bs)] {
			if {$info($node,bs) > $bootstrap} {
				set bootstrap $info($node,bs)
			}
		}
	}
	set shallowinternals [list]
	set deepinternals [list]
	foreach node $intnodes {
		if {([lsearch $extnodes [lindex $childs($node) 0]] != -1) && ([lsearch $extnodes [lindex $childs($node) 1]] != -1)} {
			lappend shallowinternals $node
		} else {
			lappend deepinternals $node
		}
	}
	set intnodes [concat $shallowinternals $deepinternals]
	set output ""
		append output "\[TOPOLOGY\]\n"
		append output "[llength $intnodes]  0\n"
		foreach node $intnodes {
			set lchild [lindex $childs($node) 0]
			set rchild [lindex $childs($node) 1]
			if [info exists info($lchild,ln)] {
				set llen [format "%1.6f" [expr $info($lchild,ln) * 100]]
			} else {
				set llen 1.000000
				puts stderr "Warning: asuming branchlength [expr $llen / 100.0] for node $lchild"
			}
			if [info exists info($rchild,ln)] {
				set rlen [format "%1.6f" [expr $info($rchild,ln) * 100]]
			} else {
				set rlen 1.000000
				puts stderr "Warning: asuming branchlength [expr $rlen / 100.0] for node $rchild"
			}
			append output "[format "% 5d" $transnode($node)][format "% 6d" $transnode($lchild)][format "% 6d" $transnode($rchild)]  $llen  $rlen\n"
		}
		set root $node
		append output "\[SEQUENCES\]\n"
		foreach node $extnodes {
			if [info exists info($node,sn)] {
				set name $info($node,sn)
			} else {
				set name "Mistery_sequence"
			}
			regsub -all "_" $name " " name
			set x [string length $name]
			while {$x < 40} {
				append name " "
				incr x
			}
			append output "[format "% 3d" $transnode($node)]   $name\n"
		}
		if $bootstrap {
			append output "\[BOOTSTRAP\]\n"
			append output "$bootstrap\n10.000000\n0\n0\n"
			foreach node $intnodes {
				if {$node == $root} {
					append output "0 0\n"
					continue
				}
				if [info exists info($node,bs)] {
					append output "$transnode($node) $info($node,bs)\n"
				} else {
					append output "$transnode($node) 0\n"
				}
			}
			append output "-13\n0\n0\n0\n700\nnoitalics\n0\nTimes New Roman\n"
		} else {
			append output "\[NOBOOTSTRAP\]\n"
		}
		append output "\[CUSTOMIZE GENERAL\]\nOTHER METHOD\nNodistance\n12632256\n15\n10.000000\n1\n1\n-13\n0\n0\n0\n700\n"
		append output "noitalics\n0\nTimes New Roman\n\[NO TRIANGLES\]\n\[NO TAXON DESIGNATIONS\]\n"
		append output "\[NO PARATAXON DESIGNATIONS\]\n\[NO SQUARES\]\n0\n0\n0\n0\n0\n\[NO ARROWS OR DOTS\]\n\[NOTREEDESCRIPTION\]\n"
	return $output
}

proc SlideRoot {tree fraction} {
	set firstlist [BreakParts $tree]
	set leftlength ""
	set rightlength ""
	set leftpart [lindex $firstlist 0]
	set rightpart [lindex $firstlist 1]
	if {$fraction > 1.0} {set fraction 1.0} elseif {$fraction < 0.0} {set fraction 0.0}
	regexp {:([0-9.]*)(\[[^\]]+\])?$}  $leftpart tmp leftlength
	if {$leftlength == ""} {set leftlength 0.0}
	regexp {:([0-9.]*)(\[[^\]]+\])?$} $rightpart tmp rightlength
	if {$rightlength == ""} {set rightlength 0.0}
	set totallength [expr $leftlength + $rightlength]
	set leftlength  [expr $totallength * $fraction]
	set rightlength [expr $totallength * (1.0 - $fraction)]
	regsub {:?[0-9.]*(\[[^\]]+\])?$}  $leftpart ":$leftlength\\1"  leftpart
	regsub {:?[0-9.]*(\[[^\]]+\])?$} $rightpart ":$rightlength\\1" rightpart
	return "(${leftpart},${rightpart});"
}

set cl $argv 
regsub -all { -} $cl {@} cl
regsub {^-} $cl "" cl
set clolist [split $cl "@"]
foreach clo $clolist {
	switch [lindex $clo 0] {
		b {
			set ADD_NODENR_AS_BOOTSTRAP 1
		}
		n {
			set DUMP_BOOTSTRAPS 1
		}
		i {
			if {[llength $clo] < 2} {
				puts stderr "No argument given for option -i"
				giveusage
				exit 1
			} else {
				set INFILENAME [lindex $clo 1]
				set USE_INFILE 1
				if ![file exists $INFILENAME] {
					puts stderr "File $INFILENAME does not exist"
					giveusage
					exit 1
				}
			}
		}
		o {
			if {[llength $clo] < 2} {
				puts stderr "No argument given for option -o"
				giveusage
				exit 1
			} else {
				set OUTFILENAME [lindex $clo 1]
				set USE_OUTFILE 1
			}
		}
		d {
			if {[llength $clo] < 2} {
				puts stderr "No argument given for option -d"
				giveusage
				exit 1
			} else {
				set RESET_DISTANCE [lindex $clo 1]
			}
		}
		s {
			if {[llength $clo] < 2} {
				puts stderr "No argument given for option -s"
				giveusage
				exit 1
			} else {
				set SCALE_DISTANCE [lindex $clo 1]
			}
		}
		l {
			if {[llength $clo] < 2} {
				puts stderr "No argument given for option -l"
				giveusage
				exit 1
			} else {
				set SLIDE_ROOT 1
				set SLIDE_FACTOR [lindex $clo 1]
			}
		}
		default {
			giveusage
			exit 1
		}
	}
}

if $USE_INFILE {
	set infile [open $INFILENAME]
	set input [read $infile]
	close $infile
} else {
	set input [read stdin]
}
if ![regexp {[^\(]*(\(.+\)[:0-9\'\". ]*;)} $input tmp tree] { ;# right brace to satisfy nedit )
	puts stderr "Error: Can't find tree"
	exit 1
}

if $SLIDE_ROOT {
	set tree [SlideRoot $tree $SLIDE_FACTOR]
}

if ![IsBifurcated $tree] {
	puts stderr "Error: Sorry, tree is not bifurcated"
	exit 1
}

set nodecnt 0
# decompose the tree
set root [DecomposeAndNumber $tree names children parents sisters nodecnt leaves]

# make list with all node numbers
set x 1; set nrnodes [expr 2 * [llength $leaves]]
set nodes [list]
while {$x < $nrnodes} {
	lappend nodes $x
	incr x
}

# make list with all node numbers of internal nodes
set intnodes [list]
foreach node $nodes {
	if ![expr [lsearch $leaves $node] + 1] {
		lappend intnodes $node
	}
}

# read additional information (lengths, bootstrap values, ...)
ReadNodeInfo names leaves info

# if NHX tree and no sequence names are given use species name (an NHX flag) as sequence name
foreach node $leaves {
	if ![info exists info($node,sn)] {
		if [info exists info($node,sp)] {
			set info($node,sn) $info($node,sp)
		}
	}
}

if $ADD_NODENR_AS_BOOTSTRAP {
	MakeNodeNrBootst $intnodes info
}
if $DUMP_BOOTSTRAPS {
	foreach node $intnodes {
		if [info exists info($node,bs)] {
			unset info($node,bs)
		}
	}
}
if $RESET_DISTANCE {
	foreach node $nodes {
		set info($node,ln) $RESET_DISTANCE
	}
}
if $SCALE_DISTANCE {
	foreach node $nodes {
		if [info exists info($node,ln)] {
			set info($node,ln) [expr $info($node,ln) * $SCALE_DISTANCE]
		}
	}
}
if {$RESET_DISTANCE || $SCALE_DISTANCE} {
	puts stderr "Warning: distances are changed !!!"
}
set treeconfile [WriteTreeconFile $intnodes $leaves children names info]

if $USE_OUTFILE {
	set outfile [open $OUTFILENAME w]
	puts $outfile $treeconfile
	close $outfile
} else {
	puts -nonewline $treeconfile
}
exit 0
